/*
 *  Copyright (c) 1990 Regents of the University of Michigan.
 *  All rights reserved.
 *
 *  show.c
 */

#ifndef lint 
static char copyright[] = "@(#) Copyright (c) 1990 Regents of the University of Michigan.\nAll rights reserved.\n";
#endif

#include "dixie.h"
#include <stdio.h>
#include <string.h>

#if !defined(__STDC__) && !defined(VMS)
#include <memory.h>
#endif

#ifdef DOS
#include <malloc.h>
#include "hostform.h"
#endif

#ifdef MACOS
#include <types.h>
#include "tcp.h"
#endif MACOS

#if !defined(MACOS) && !defined(DOS)
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#endif MACOS

#ifndef INADDR_LOOPBACK
#define INADDR_LOOPBACK 0x7f000001
#endif

u_long *lookup_addrs();


/*
 * dxi_init - initialize the dixie library.  host is used as the dixie
 * host, timeout and retries as the default initial number of seconds
 * to wait before retransmitting packets, and number of retries to make,
 * respectively.  Supplying 0 for timeout or retries resorts to the
 * default values of 2 and 4, respectively.  Supplying NULL for the host
 * causes the localhost to be used.
 *
 * Example:
 *	dxi_init( "some.host.some.domain", 2, 4 )
 */

dxi_init( host, timeout, retries )
char	*host;
int	timeout;
int	retries;
{
#if !defined( DOS ) && !defined( MACOS )
	struct servent *sp;
	extern void attempt_rebind();
#endif DOS
    	u_long		ipaddr;
	extern char	*dxi_buffer;
	extern int	dxi_timeout, dxi_retries;
	extern u_long	*dixieaddr;
	extern u_short	dixieport;

	dxi_errno = DXI_ERROR;	/* pessimistic, but efficient... */
	if ( dxi_buffer == NULL ) {
		if ( (dxi_buffer = (char *) malloc( MAX_PKT_SIZE )) == NULL )
			return( -1 );
	}

	if ( dixieaddr != NULL ) {
	    free( dixieaddr );
	}

#ifdef DOS
	ncsainit();
#endif DOS

	if ( host == NULL || ( ipaddr = inet_addr( host )) != -1 ) {
	    if (( dixieaddr = ( u_long *)malloc( 2 * sizeof( u_long )))
		    == NULL ) {
#ifdef DOS
		netshut();
#endif DOS
		return( -1 );
	    }

	    if ( host == NULL ) {
		dixieaddr[ 0 ] = htonl( INADDR_LOOPBACK );
	    } else {
		dixieaddr[ 0 ] = ipaddr;
	    }
	    dixieaddr[ 1 ] = 0L;
	} else {
	    if (( dixieaddr = lookup_addrs( host )) == NULL ) {
#ifdef DOS
		netshut();
#endif DOS
		return( -1 );
	    }
	}

#if defined(MACOS) || defined(DOS)
	dixieport = htons( DXI_PORT );
#else MACOS
	if (( sp = getservbyname( "dixie", "tcp" )) == NULL ) {
	    dixieport = htons( DXI_PORT );
	} else {
	    dixieport = htons( sp->s_port );
	}

	(void) signal( SIGPIPE, attempt_rebind );
#endif

	dxi_timeout = timeout <= 0 ? 2 : timeout;
	dxi_retries = retries <= 0 ? 4 : retries;
	dxi_options = 0;

	dxi_errno = DXI_SUCCESS;
	return( 0 );
}


#if !defined( MACOS ) && !defined( DOS )
u_long *
lookup_addrs( host )
    char	*host;
{
    struct hostent	*hp;
    int			numaddrs;
    u_long		*addrs;

    if ( (hp = gethostbyname( host )) == NULL ) {
	return( NULL );
    }

#ifdef h_addr
    for ( numaddrs = 0; hp->h_addr_list[ numaddrs ] != 0; ++numaddrs ) {
	;
    }
#else h_addr
    numaddrs = 1;
#endif h_addr

    if (( addrs = (u_long *)malloc( ( numaddrs + 1 ) * sizeof( u_long )))
	    == NULL ) {
	return( NULL );
    }
    addrs[ numaddrs ] = 0;

#ifdef h_addr
    while ( --numaddrs >= 0 ) {
	(void) memcpy( (char *)&addrs[ numaddrs ],
		(char *)hp->h_addr_list[ numaddrs ], sizeof( u_long ));
    }
#else h_addr
    (void) memcpy( (char *)&addrs[0] , (char *)&hp->h_addr, sizeof( u_long ));
#endif h_addr

    return( addrs );
}
#endif


#ifdef MACOS
u_long *
lookup_addrs( host )
    char	*host;
{
    struct hostInfo	*hip;
    int			numaddrs;
    u_long		*addrs;

    if (( hip = gethostinfobyname( host )) == NULL ) {
	return( NULL );
    }
    
    for ( numaddrs = 0; hip->addr[ numaddrs ] != 0; ++numaddrs ) {
	;
    }

    if (( addrs = (u_long *)malloc( ( numaddrs + 1 ) * sizeof( u_long )))
	    == NULL ) {
	return( NULL );
    }
    addrs[ numaddrs ] = 0;

    while ( --numaddrs >= 0 ) {
	(void) memcpy( (char *)&addrs[ numaddrs ],
		(char *)&hip->addr[ numaddrs ], sizeof( u_long ));
    }

    return( addrs );
}
#endif MACOS


#ifdef DOS
u_long *
lookup_addrs( host )
    char	*host;
{
    struct machinfo	*mr;
    int			numaddrs, i;
    char		*ipp;
    u_long		*addrs, along;

    if (( mr = lookup( host )) == NULL ) {
	return( NULL );
    }

    ipp = mr->hostip;
#ifdef NCSADOMAINFIX
    numaddrs = 0;
    while ( numaddrs < 4 ) {	/* maximum of 4 addresses */
        (void) memcpy( (char *)&along, (char *)ipp, sizeof( u_long ));
	if ( along == 0 ) {
		break;
	}
	++numaddrs;
	ipp += 4;
    }
#else /* NCSADOMAINFIX */
    numaddrs = 1;
#endif /* NCSADOMAINFIX */

    if (( addrs = (u_long *)malloc(( numaddrs + 1 ) * sizeof( u_long )))
		== NULL ) {
	return( NULL );
    }
    addrs[ numaddrs ] = 0;

    for ( i = 0, ipp = mr->hostip; i < numaddrs; ++i, ipp += 4 ) {
    	(void) memcpy( (char *)&addrs[ i ], (char *)ipp, sizeof( u_long ));
    }

    return( addrs );
}
#endif DOS
