/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include "dixie.h"

#include <quipu/attrvalue.h>
#include <quipu/ds_error.h>
#include <quipu/commonarg.h>
#include <quipu/compare.h>
#include <quipu/dua.h>

#define NULL	0

#define COMPARE_END(x)	{ dn_free(thedn); \
			free(buff); \
			free(value); \
			return(x); }

extern int	debug;
extern int	bound;
extern int	stream_connected;
extern short	timelimit, sizelimit;


/* args for the compare call */
struct ds_compare_arg	compare_arg = { default_common_args };
struct ds_compare_result compare_result;
struct DSError		compare_error;

compareop(bp, rp, dontderef )
char	*bp, *rp;
char	dontderef;
{
	DN		thedn = NULLDN;
	char		buff[1024];
	int		flag;
	int		i;
	int		limit;
	PS		ps;
	int		len;
	char	        *p, *value;


	compare_arg.cma_common.ca_servicecontrol.svc_timelimit = timelimit;
	compare_arg.cma_common.ca_servicecontrol.svc_sizelimit = sizelimit;
	if ( dontderef )
		compare_arg.cma_common.ca_servicecontrol.svc_options |=
		    SVC_OPT_DONTDEREFERENCEALIAS;
	else
		compare_arg.cma_common.ca_servicecontrol.svc_options &=
		    ~SVC_OPT_DONTDEREFERENCEALIAS;
	*rp = '\0';
	if ( debug & 2 ) printf("Compare (%s)\n", bp);

	/* 
	 * First convert the dn given into a DN structure
	 */

	if ( *bp == '@') bp++;
	if ( (thedn = str2dn(bp)) == NULLDN ) {
		if ( debug > 1 ) printf("compareops: str2dn (%s) failed\n",bp);
		COMPARE_END(DXI_BADDN);
	}
	compare_arg.cma_object = thedn;

	bp += strlen(bp) + 1;	/* skip the dn */

	/* get the attrs and values to modify */
	if ( debug ) printf("getting attribute and value...\n");
	strcpy(buff, bp);
	if ( debug ) printf("\t(%s)\n", buff);
	p = bp;
	p += strlen(bp);
	p += 1;
	value = malloc(strlen(p) + 1);
	strcpy(value, p);
	if ( debug ) printf("\t(%s)\n", value);
	if ( debug ) printf("invoking get_ava in server\n");
	fflush(stdout);
	get_ava(&compare_arg.cma_purported, buff, value);
	if ( debug) printf("Invoking ds_compare....\n");
	if ( ds_compare(&compare_arg, &compare_error, &compare_result) != DS_OK )
	  if ( debug )
	    printf("ds_compare failed\n");
	else
	  if ( debug )
		printf("compare successful - checking for errors\n");
	
	switch ( compare_error.dse_type ) {
	case DSE_REMOTEERROR:
		 if ( debug ) {
			printf("\tservice error\n");
			error( &compare_error );
		      }
		COMPARE_END(DXI_DSA_DOWN);
	case DSE_REFERRAL:
		 if ( debug ) {
			printf("\treferral error\n");
			error( &compare_error );
		      }
		COMPARE_END(DXI_REFER_ERROR);
	case DSE_NAMEERROR:
		 if ( debug ) {
			printf("\tname error\n");
			error( &compare_error );
		      }
		COMPARE_END(DXI_NAME_ERROR);
	case DSE_SERVICEERROR:
		 if ( debug ) {
			printf("\tservice error\n");
			error( &compare_error );
		      }
		if (compare_error.dse_un.dse_un_service.DSE_sv_problem != DSE_SV_TIMELIMITEXCEEDED)
			COMPARE_END(DXI_ERROR)
		else
		        COMPARE_END(DXI_LIMIT);
	case DSE_ATTRIBUTEERROR:
		 if ( debug ) {
			printf("\tattribute error\n");
			error( &compare_error );
		      }
		COMPARE_END(DXI_NEX_ATTR);
	case DSE_NOERROR:
		if ( debug ) printf("\tno error\n");
		if ( compare_result.cmr_matched == TRUE ) {
		  if ( debug ) printf("\tmatched\n");
		  COMPARE_END(DXI_SUCCESS);
		}
		else {
		  if (compare_result.cmr_matched == FALSE ) {
		    if ( debug) printf("\tmatch failed\n");
		    COMPARE_END(DXI_COMPAREFAILED);
		  }
		}
	default:
		if ( debug) {
		  printf("\tsome unknown error (%d)\n",
		    compare_error.dse_type);
		  error( &compare_error );
		}
		COMPARE_END(DXI_ERROR);
	}
}

get_ava (avasert, type, value)
AVA            *avasert;
char           *type;
char           *value;
{
	char           *TidyString ();

	if (type == NULLCP || *type == 0)
		return (NOTOK);

	avasert->ava_type = AttrT_new (TidyString (type));
	if (avasert->ava_type == NULLAttrT) {
		printf ("Invalid at %s\n", type);
		return (NOTOK);
	}

	if ((avasert->ava_value = str_at2AttrV (TidyString (value), avasert->ava_type)) == NULLAttrV) {
		printf("Invalid attribute value value\n");
		AttrT_free (avasert->ava_type);
		return (NOTOK);
	}

	return (OK);
}
