/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "dixie.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/syslog.h>
#include <sys/wait.h>
#include <signal.h>
#include "packet.h"
#include <quipu/ds_error.h>
#include <quipu/attrvalue.h>
#include <quipu/bind.h>

#ifndef FD_SET
#define NFDBITS         32
#define FD_SETSIZE      32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

int	debug;
int	dosyslog;
int	stream_connected;
int	bound;
int	port = -1;
int	udp, tcp;
short	timelimit, sizelimit;
int	inactive_secs = 12 * 3600;
long	stream_conn_start_time;

static int noactivity();

char	result[MAX_PKT_SIZE];
char	**Argv;
int	Argc;

main (argc, argv)
int	argc;
char	**argv;
{
	char			*bp, *rp = result + DXI_H_SIZE;
	char			buf[MAX_PKT_SIZE];
	int			s2, ns;
	long			len, pktsize;
	u_long			netlong; /* andy */
	u_short			netshort;
	int			nbytes, rc, i, rplen;
	short			seq, id;
	struct sockaddr_in	from;
	struct timeval		tv;
	int			fromlen;
	long			totlen, tosend, totsend;
	long			rtotlen;
	short			sequence;
	char			outbuf[MAX_PKT_FRAG + DXI_H_SIZE];
	char			*dp;
	int			dsapargc;
	char			**dsapargv;
	int			wait4child();
	Packet			*make_pkt();
	extern char		*optarg;

	if ((dsapargv = (char **) malloc(4 * sizeof(char *))) == NULL) {
		perror("malloc");
		clean_exit(1);
	}
	dsapargv[0] = argv[0];
	dsapargv[1] = 0;
	dsapargv[2] = 0;
	dsapargv[3] = 0;
	dsapargc = 1;

	/* use getopt because Bryan is omnipotent */
	while ((i = getopt(argc, argv, "c:d:p:t:l")) != EOF)
		switch(i) {
		case 'p' :
			port = atoi(optarg);
			break;

		case 'l' :
			dosyslog = 1;
			break;

		case 'd' :
			debug = atoi(optarg);
			break;

		case 'c' :
			dsapargv[1] = "-call";
			dsapargv[2] = strdup(optarg);
			dsapargc = 3;
			break;

		case 't' :
			inactive_secs = 3600 * atoi(optarg);
			break;

		default :
			printf("Invalid arg! %c\n", i);
			usage(argv[0]);
		}

	/* detach if stderr is redirected or no debugging */
	(void) detach(debug);

	/* if syslogging requested, initialize */
	if (( bp = rindex( argv[ 0 ], '/' )) == NULL ) {
	    bp = strdup( argv[ 0 ] );
	} else {
	    bp = strdup( bp + 1 );
	}
	    
	if (dosyslog && openlog(bp, LOG_PID | LOG_NOWAIT, LOG_LOCAL1 ) < 0) {
	    /*
	    perror("openlog");
	    clean_exit(-1);
	    */
	}
	if (dosyslog) syslog(LOG_INFO|LOG_LOCAL1, "initializing" );
	Argv = argv;		/* for setproctitle()'s use */
	Argc = argc;
	setproctitle( "initializing" );

        /* load the syntax handlers for asn.1 types quipu knows about */
        (void) quipu_syntaxes();

	(void) signal(SIGCHLD, wait4child);

        /* load oid tables, set dsa name, etc... */
        (void) dsap_init(&dsapargc, &dsapargv);
	get_syntaxes();

	/* bind to the directory */
	(void) nullbind();

	/* set up the socket to listen on */
	(void) set_socket();

	if (dosyslog) syslog(LOG_INFO|LOG_LOCAL1, "awaiting requests" );
	if ( port == -1 ) {
	    setproctitle( "awaiting requests" );
	} else {
	    setproctitle( "p%d awaiting requests", port );
	}

	/* read names and look them up in the directory */
	for (;;) {
		fflush( stdout );
		fromlen = sizeof(from);
		if ( (nbytes = get( buf, sizeof(buf), &from, &fromlen ))
		    < 0 ) {
			if ( debug ) perror("get returns -1");
			if ( stream_connected ) {
				ds_unbind();
				clean_exit(1);
			} else
				continue;
		} else if ( nbytes == 0 ) {
			if ( debug & 4 )
				printf("tcp bind request successful\n");
			continue;
		}

		if ( debug & 4 ) {
			printf("got this (%d bytes):\n",nbytes);
			bprint(buf, nbytes);
		}

		if ( bound == 0 )
			(void) nullbind();

		/*
		 * Check for duplicate packet.  A history of packets
		 * is kept.  Each packet is held for at least 60 seconds,
		 * packets are expired when new ones are inserted.  acks
		 * and nacks are neither checked for duplicates
		 * nor inserted into the packet queue.
		 */

		if (*buf != DXI_ACK && *buf != DXI_NACK) {
			if (pkt_insert(make_pkt(buf, &from, nbytes, 0))
			    == PKT_DUPLICATE) {
				if ( debug ) 
					printf("duplicate packet\n");

				/*
				 * if we found the packet, pkt_resend will
				 * arrange to have it sent.  If not, we must
				 * retry the operation (should not happen
				 * since we just found it above...
				 */

				if (pkt_resend(buf, &from, rp) == DXI_SUCCESS)
					*buf = DXI_RESEND;
			}
		}

		/* 
		 * Break out the arguments and query the directory.
		 * The result of the query, ready to be sent back, will be
		 * left in the global buffer result.
		 */

		rtotlen = seq = id = -1;
		bp = buf + DXI_H_SIZE;	/* skip header */
		bcopy( &buf[DXI_H_TIMELIMIT], (char *) &netshort,
		    sizeof(short) );
		timelimit = ntohs(netshort);
		if ( timelimit == 0 )
			timelimit = SVC_NOTIMELIMIT;
		bcopy( &buf[DXI_H_SIZELIMIT], (char *) &netshort,
		    sizeof(short) );
		sizelimit = ntohs(netshort);
		if ( buf[DXI_H_VERSION] == 0 ) {
			/* only one byte for returned count! */
			if ( sizelimit == 0 || sizelimit > 255 ) {
				sizelimit = 255;
			}
		}
		if ( sizelimit == 0 )
			sizelimit = SVC_NOSIZELIMIT;

		switch (*buf) {

		case DXI_PING:		/* "Are you there?" request */
			if ( debug ) printf("Ping...\n");
			*rp = 0;
			rplen = 0;
			rc = pingop(bp, nbytes - DXI_H_SIZE, rp, &rplen, from);
			break;
		case DXI_ACK:		/* packet acknowledgement */
			if ( debug ) printf("Ack...\n");
			if ( pkt_remove(make_pkt(buf, &from, nbytes))
			    == PKT_NOTFOUND )
				if ( debug ) printf("packet not found\n");
			continue;
			/* NOTREACHED */
			break;

		case DXI_NACK:		/* negative packet acknowledgement */
			if ( debug ) printf("Nack...\n");
			rc = pkt_nack(buf, &from, rp, &seq, &id, &rtotlen);
			break;

		case DXI_RESEND:
			if ( debug ) printf("Resend...\n");
			break;

		case DXI_READ:		/* read request */
			if ( debug ) printf("Read...\n");
			rplen = sizeof( result );
			rc = readop(bp, rp, &rplen, buf[ DXI_H_OPTIONS ]
			    & DXI_OPT_FILEATTR, buf[ DXI_H_OPTIONS ]
			    & DXI_OPT_DONTDEREF, buf[DXI_H_VERSION] );
			break;

		case DXI_LIST:		/* list request */
			if ( debug ) printf("List...\n");
			rplen = sizeof( result );
			rc = listop(bp, rp, rplen, buf[DXI_H_OPTIONS]
			    & DXI_OPT_DONTDEREF);
			rplen = strlen(rp + 2) + 2;
			break;

		case DXI_ADD:		/* add request */
			if ( debug ) printf("Add...\n");
			rplen = sizeof( result );
			if ( stream_connected ) {
				rc = addop(bp, rp, rplen, buf[DXI_H_OPTIONS]
				    & DXI_OPT_DONTDEREF);
			} else {
				*rp = '\0';
				rc = DXI_NOTCONN;
			}
			rplen = strlen(rp);
			break;

		case DXI_REMOVE:	/* remove request */
			if ( debug ) printf("Remove...\n");
			rplen = sizeof( result );
			if ( stream_connected ) {
				rc = removeop(bp, rp, rplen, buf[DXI_H_OPTIONS]
				    & DXI_OPT_DONTDEREF);
			} else {
				*rp = '\0';
				rc = DXI_NOTCONN;
			}
			rplen = strlen(rp);
			break;

		case DXI_MODRDN:	/* modifyrdn request */
			if ( debug ) printf("Modrdn...\n");
			rplen = sizeof( result );
			if ( stream_connected ) {
				rc = modrdnop(bp, rp, rplen, buf[DXI_H_OPTIONS]
				    & DXI_OPT_DONTDEREF);
			} else {
				*rp = '\0';
				rc = DXI_NOTCONN;
			}
			rplen = strlen(rp);
			break;

		case DXI_SEARCH:	/* search request */
		case DXI_SEARCH1:
		case DXI_SEARCH_APPROX:
		case DXI_SEARCH_EXACT:
		case DXI_SEARCH_ISUB:
		case DXI_SEARCH_FSUB:
		case DXI_SEARCH_IFSUB:
		case DXI_SEARCH_GEN:
			if ( debug ) printf("Search...\n");
			rplen = sizeof( result );
			rc = searchop(bp, rp, *buf, &rplen, buf,
			    buf[DXI_H_OPTIONS] & DXI_OPT_DONTDEREF,
			    buf[DXI_H_OPTIONS] & DXI_OPT_DONTSEARCHALIASES );
			break;

		case DXI_WRITE:		/* modify requet */
			if ( debug )
				printf("Write...%d\n", buf[DXI_H_OPTIONS] &
				    DXI_OPT_DONTDEREF);	/* andy */
			if ( stream_connected )
				rc = writeop(bp, rp,
				    buf[DXI_H_OPTIONS] & DXI_OPT_DONTDEREF);
			else {
				*rp = '\0';
				rc = DXI_NOTCONN;
			}
			rplen = strlen( rp );
			break;

		/* bind request */
		case DXI_BIND:
			if ( debug ) printf("Bind...\n");

			/* if non-null DN, check the pw first */
			if ( *bp != '\0' && (rc = pwcompare( bp )) != 0 ) {
				*rp = '\0';
				rplen = 0;
				break;
			}

			/* handle binds when already bound */
			if ( stream_connected ) {
				rc = bbindop(bp, rp, 1);
				rplen = strlen( rp );
				break;
			}

			/* handle binds when not already bound */
			if ( (rc = bindop(bp, rp, &s2)) == -1 ) {

				/* could not fork -- report error */
				if ( debug ) perror("parent: bad fork");
				rc = DXI_ERROR;

			} else if ( rc == 0 ) {

				/* parent: forked ok -- let child send ack */
				if ( debug ) printf("parent: forked ok\n");
				continue;

			} else {

				/* child: send ack, wait for a connection */
				if ( debug ) printf("child: acking\n");

				result[DXI_H_RC] = rc;
				bcopy( &buf[DXI_H_ID], &result[DXI_H_ID],
				    sizeof(short));
				len = DXI_H_SIZE + 1 +
				    strlen(&result[DXI_H_SIZE]);
				pktsize = len - DXI_H_SIZE;
				netlong = htonl( pktsize ); /* andy */
				bcopy((char *) &netlong, &result[DXI_H_TOTLEN],
				    sizeof(long));
				fromlen = sizeof(from);
				if ( debug & 4 ) {
					printf("bind sending (%d)\n", len);
					bprint(result, len);
				}
				pkt_insert(make_pkt(result, &from, pktsize,
				    1));
				if ( send(udp, result, len, &from, fromlen) 
				    < 0 ) {
					if ( debug ) perror("child: bad send");
					clean_exit(1);
				}

				/* something wrong with the bind */
				if ( rc != DXI_SUCCESS ) {
					if ( debug ) 
						printf("child: bad bind\n");
					clean_exit(1);
				}
				listen(s2, 1);
				fromlen = sizeof(from);
				if ( debug & 2 ) printf("child: waiting...\n");
				if ( (ns = accept(s2, &from, &fromlen)) < 0 ) {
					if ( debug ) 
						perror("child: bad accept");
					clean_exit(1);
				}
				if ( debug ) printf("connected\n");
				close(s2);
				dup2(ns, tcp);
				close(ns);
				udp = -1;
				stream_connected = 1;
				gettimeofday( &tv, (struct timezone *)NULL );
				stream_conn_start_time = tv.tv_sec;
				if ( inactive_secs > 0 ) {
				    signal( SIGALRM, noactivity );
				    alarm( inactive_secs );
				}
				continue;
			}
			break;

		case DXI_COMPARE:	/* compare requet */
			if ( debug )
				printf("Compare...%d\n", buf[DXI_H_OPTIONS] &
				    DXI_OPT_DONTDEREF);	/* andy */
			rc = compareop(bp, rp,
			    buf[DXI_H_OPTIONS] & DXI_OPT_DONTDEREF);
			rplen = strlen( rp );
			break;

		/* invalid request -- ignore it */
		default:
			if ( debug ) printf("invalid request (%d)\n", *buf);
			result[DXI_H_SIZE] = '\0';
			rc = DXI_BADCMD;
			rplen = 0;
			break;
		}

		result[DXI_H_RC] = rc;
		result[DXI_H_VERSION] = buf[DXI_H_VERSION];
		if (id != -1) {
			netshort = htons( id );
			bcopy((char *)&netshort, &result[DXI_H_ID],
			    sizeof(short));
		} else
			bcopy(&buf[DXI_H_ID], &result[DXI_H_ID],
			    sizeof(short));
		totlen = DXI_H_SIZE + 1 + rplen;
		if (id != -1 && (totlen - DXI_H_SIZE) > MAX_PKT_FRAG)
			totlen--;
		tosend = totsend = totlen - DXI_H_SIZE;
		if (rtotlen == -1) {
			netlong = htonl( totsend ); /* andy */
			bcopy((char *) &netlong, &result[DXI_H_TOTLEN],
			    sizeof(long));
		} else {
			netlong = htonl( rtotlen ); /* andy */
			bcopy((char *) &netlong, &result[DXI_H_TOTLEN],
			    sizeof(long));
		}

		if ( debug & 4 ) {
			printf("total packet is this (%d bytes):\n", totlen);
			if (stream_connected || result[DXI_H_VERSION] >= 1)
				bprint(result, totlen);
		}

		if (stream_connected) {
			if (send(tcp, result, totlen, &from, fromlen) < 0) {
				if ( debug ) perror("send");
				clean_exit(1);
			}
			continue;
		}
		if ( result[DXI_H_VERSION] >= 1 ) {
			if ( debug & 4 )
				printf("UDP: sending single packet\n");
			if (send( udp, result, totlen, &from, fromlen ) < 0) {
				if ( debug ) perror("send");
			}
			continue;
		}

		if (id == -1 && *buf != DXI_RESEND)
			pkt_insert(make_pkt(result, &from, totlen, 1));

		if ( debug & 4 )
			printf("UDP: fragmenting\n");
		dp = &result[DXI_H_SIZE];
		bcopy(result, outbuf, DXI_H_SIZE);
		sequence = 0;
		while ( tosend > 0 ) {
			pktsize = tosend > MAX_PKT_FRAG ? MAX_PKT_FRAG:tosend;
			bcopy(dp, outbuf + DXI_H_SIZE, pktsize);
			if (seq != -1) {
				sequence = seq;
			}
			netshort = htons(sequence);
			bcopy((char *) &netshort,
			    &outbuf[DXI_H_SEQUENCE], sizeof(short));

			if ( debug & 4 ) {
				printf("fragment %d is (%d bytes):\n",
				    sequence, pktsize + DXI_H_SIZE);
				bprint(outbuf, pktsize + DXI_H_SIZE);
			}

			if ( send(udp, outbuf, pktsize + DXI_H_SIZE, &from,
			    fromlen) < 0 )
				if ( debug ) perror("send");

			if ( seq != -1 ) {	/* sent a NACK */
			    break;
			}
			sequence++;
			tosend -= pktsize;
			dp += pktsize;
		}
	}	
}


static set_socket()
{
	struct sockaddr_in	addr;
	struct servent		*serv;
	int			tcpport, udpport;

	if ( port == -1 ) {
		if ( (serv = getservbyname( "dixie", "tcp" )) == NULL )
			tcpport = DXI_PORT;
		else
			tcpport = serv->s_port;
		if ( (serv = getservbyname( "dixie", "udp" )) == NULL )
			udpport = DXI_PORT;
		else
			udpport = serv->s_port;
	} else {
		tcpport = port;
		udpport = port;
	}

	if ( (udp = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ) {
		perror("socket");
		clean_exit(1);
	}

	/* bind to a name */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(udpport);
	if ( bind(udp, &addr, sizeof(addr)) ) {
		perror("bind");
		clean_exit(1);
	}
	if ( debug ) printf("udp socket allocated and bound\n");
	if ( (tcp = socket(AF_INET, SOCK_STREAM, 0)) == -1 ) {
		perror("socket");
		clean_exit(1);
	}

#ifndef AUX
	/* set option so clients can't keep us from coming back up */
	if ( setsockopt( tcp, SOL_SOCKET, SO_REUSEADDR, 0, 0 ) < 0 ) {
		perror( "setsockopt" );
		clean_exit(1);
	}
#endif

	/* bind to a name */
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(tcpport);
	if ( bind(tcp, &addr, sizeof(addr)) ) {
		perror("bind");
		clean_exit(1);
	}
	if ( listen( tcp, 5 ) == -1 ) {
		perror( "listen" );
		clean_exit( 1 );
	}
	if ( debug ) printf("tcp socket allocated, bound, and listening\n");
}


static get(b, blen, from, fromlen)
char			*b;
int			blen;
struct sockaddr_in	*from;
int			*fromlen;
{
	int			n, ns;
	long			totlen, gotlen;
	fd_set			readfds;
	struct hostent		*hp;
	struct timeval		tv;

	FD_ZERO( &readfds );
	if ( tcp >= 0 )
		FD_SET( tcp, &readfds );
	if ( udp >= 0 )
		FD_SET( udp, &readfds );

	if ( select( getdtablesize(), &readfds, 0, 0, 0 ) == -1 )
		return( -1 );

	if ( stream_connected ) {
		if ( inactive_secs > 0 ) {
		    alarm( inactive_secs );
		}
		totlen = -1;
		gotlen = 0;
		do {
		    if ( (n = read( tcp, b+gotlen, blen-gotlen )) <= 0 ) {
			if ( debug ) printf("read returns -1\n");
			return( -1 );
		    }
		    gotlen += n;
		    if ( totlen == -1 &&
				gotlen >= DXI_H_TOTLEN + sizeof(long)) {
			bcopy( &b[DXI_H_TOTLEN],(char *)&totlen, sizeof(long));
			if (( totlen = ntohl( totlen )) == 0 ) {
			    totlen = n; /* kludge for broken dixie clients */
			}
		    }
		} while ( totlen == -1 || gotlen < totlen );
		n = gotlen;

	} else if ( FD_ISSET( tcp, &readfds ) ) {
		ds_unbind();
		if ( (ns = accept( tcp, from, fromlen )) == -1 ) {
			if ( debug ) perror( "accept" );
			clean_exit( 1 );
		}

		switch ( fork() ) {
		case 0:		/* child */
			close( tcp );
			close( udp );
			tcp = ns;
			udp = -1;
			stream_connected = 1;
			gettimeofday( &tv, (struct timezone *)NULL );
			stream_conn_start_time = tv.tv_sec;
			if ( inactive_secs > 0 ) {
			    signal( SIGALRM, noactivity );
			    alarm( inactive_secs );
			}
			hp = gethostbyaddr((char *) &(from->sin_addr.s_addr),
			    sizeof(from->sin_addr.s_addr), AF_INET);
			syslog( LOG_INFO, "TCP connection from %s (%s)",
			    ( hp == NULL ) ? "unknown" : hp->h_name,
			    inet_ntoa( from->sin_addr ));
			setproctitle( ( hp == NULL ) ?
			    inet_ntoa( from->sin_addr ) : hp->h_name );
			nullbind();
			return( 0 );

		case -1:	/* error */
			if ( debug )
				printf("fork failed!!\n");
			nullbind();
			return( -1 );

		default:	/* parent */
			close( ns );
			nullbind();
			return( -1 );
		}
	} else {
		if ( (n = recvfrom(udp, b, blen, 0, from, fromlen)) == -1 ) {
			if ( debug ) printf("rcvfrom returns -1\n");
			return(-1);
		}
	}

	if ( n < 3 ) return(-1);	/* malformed packet */

	b[n] = '\0';			/* should come this way, but... */
	b[n+1] = '\0';

	return(n);
}


static send(s, b, blen, from, fromlen)
int			s;
char			*b;
int			blen;
struct sockaddr_in	*from;
int			fromlen;
{
	int	n;

	if ( stream_connected ) {
		if ( (n = write(s, b, blen)) != blen ) {
			if ( debug ) printf("wrote (%d) of (%d)\n", n, blen);
			return(-1);
		}
	} else {
		if ( sendto(s, b, blen, 0, from, fromlen) == -1 ) {
			if ( debug ) {
				printf( "tried to sendto %d bytes\n", blen );
				perror("sendto");
			}
			return(-1);
		}
	}
	return(0);
}

static usage(s)
char *s;
{
	printf("Usage:  %s [-c DSA] [-d debuglevel] [-l] [-p port] [-t inactivity timeout\n", s);
	clean_exit(-1);
}

static wait4child()
{
#if defined(SunOS) && SunOS < 40
	union wait status;
#else
	int	status;
#endif

	if (debug & 8) printf("parent: catching child status\n");
	while ( wait3( &status, WNOHANG | WUNTRACED, 0 ) > 0 )
		;	/* NULL */
}


static int
noactivity()
{
	if (dosyslog) {
	    syslog(LOG_INFO|LOG_LOCAL1,
		"shutting down after %d hours of inactivity",
		inactive_secs / 3600 );
	}
	clean_exit( 1 );
}
