/*
 * Copyright (c) 1991 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dixie.h>
#ifndef __STDC__
#include <memory.h>
#endif
#include <sys/types.h>
#include "ud.h"

#define isT61(c)        (!isascii(c) || !isalnum(c) \
                        && c != 047 && c != '(' && c != ')' \
                        && c != '+' && c != '-' && c != '.' && c != ',' \
                        && c != '/' && c != ':' && c != '=' && c != '?' \
                        && c != ' ')

extern struct entry Person; 
extern int verbose;

modify(who)
char *who;
{
#ifdef UOFM
	void set_updates();	/* routine to modify noBatchUpdates */
#endif
	void set_pw();		/* routine to modify a password */
	extern char *find();	/* for finding the person to modify */
	extern char bound_as[]; /* who this user is bound as */
	extern char me[];	/* set if user has used I AM */
	char *result;		/* used to point to various results */
	char *cp;		/* miscellaneous char pointer */
	char *dn;		/* distinguished name */
	char name[BUFSIZ];	/* entry to modify */
	static char ans[RESP_SIZE];	/* for holding user input */
#ifdef UOFM
	static char printed_warning = 0;	/* for use with the */
	extern char * fetch_update_setting();	/* noBatchUpdates attr */
#endif

	/*
	 *  First, decide what entry we are going to modify.  If the
	 *  user has not included a name on the modify command line,
	 *  we will use the person who was last looked up with a find
	 *  command.  If there is no value there either, we don't know
	 *  who to modify.
	 *
	 *  Once we know who to modify, be sure that they exist, and
	 *  parse out their DN.
	 */
	if ((who == NULL) && (Person.name[0] == '\0') && !strcmp(bound_as, "Nobody") && (me[0] == '\0')) {
		printf("  Modify who? ");
		fflush(stdout);
		fetch_buffer(name, sizeof(name), stdin);
		if (name[0] != '\0')
			who = name;
		else
			return;
	}
	else if ((who == NULL) && (Person.name[0] != '\0'))
		who = Person.name;
	else if ((who == NULL) && (me[0] != '\0'))
		who = me;
	else if ((who == NULL) && strcmp(bound_as, "Nobody"))
		who = bound_as;
	if ((result = find(who)) == NULL) {
		return;
	}
	dn = dxi_getdn(dxi_firstentry(result));
	cp = strrchr(dn, '=');
	if (cp == NULL) {
		fprintf(stderr, "  Malformed distinguished name \"%s\"\n", dn);
		return;
	}
	cp++;
	if (dxi_debug)
		printf("  Modifying entry for \"%s\"\n", cp);

	/* bind if necessary */
	if (!strcmp(bound_as, "Nobody")) {
		if (auth(who, who) < 0)
			return;
	}

#ifdef UOFM
	/*
	 *  If verbose mode is turned on and the user has not set a value
	 *  for noBatchUpdates, warn them that what they are about to do
	 *  may be overwritten automatically by that Stinkbug.
	 */
	(void) fetch_update_setting(dn);
	if (verbose && !printed_warning && (dxi_errno == DXI_NEX_ATTR)) {
		printed_warning = 1;
		printf("\n  WARNING!\n");
		printf("  You are about to make a modification to an X.500 entry\n");
		printf("  that has its \"automatic updates\" field set to ON.\n");
		printf("  This means that the entry will be automatically updated\n");
		printf("  each month from official University sources like the\n");
		printf("  Personnel Office.  With \"automatic updates\" set to ON,\n");
		printf("  the following fields will be overwritten each month:\n");
		printf("         Title, home address and phone,\n");
		printf("         business address and phone\n");
		printf("  If you modify any of these fields, you may want to change\n");
		printf("  the \"automatic updates\" field to OFF so that your\n");
		printf("  changes will not be overwritten.  You may change this\n");
		printf("  setting by choosing \"u\" at the \"Modify what?\" prompt\n");
	}
#endif

	/*
	 *  Current values for user 'who' are being held in 'result'. We
	 *  should parse up that buffer and fill in the Person structure.
	 *  Once we're done with that, we can find out which fields the
	 *  user would like to modify.
	 */
	parse_answer(result);
	if (verbose) {
		printf("  You now need to specify what field you'd like to modify.\n\n");
			printf("  Choices are:\n");
			printf("  -----------------------\n");
			printf("  e  E-mail address\n");
			printf("  t  Title\n");
			printf("  p  Home phone\n");
			printf("  P  Business phone\n");
			printf("  a  Home address\n");
			printf("  A  Business address\n");
			printf("  f  Fax number\n");
			printf("  d  Description\n");
			printf("  x  Password\n");
#ifdef UOFM
			printf("  u  Automatic updates\n");
#endif
			printf("  m  Member list (for groups)\n");
			printf("\n  Hitting RETURN will abort the process.\n\n");
	}
again:
	printf("  Modify what? ");
	fflush(stdout);
	fetch_buffer(ans, sizeof(ans), stdin);
	switch (ans[0]) {
	case 'h' :
	case 'H' :
	case '?' :
		printf("\n  Choices are:\n");
		printf("  -----------------------\n");
		printf("  e  E-mail address\n");
		printf("  t  Title\n");
		printf("  p  Home phone\n");
		printf("  P  Business phone\n");
		printf("  a  Home address\n");
		printf("  A  Business address\n");
		printf("  f  Fax number\n");
		printf("  d  Description\n");
		printf("  x  Password\n");
#ifdef UOFM
		printf("  u  Automatic updates\n");
#endif
		printf("  m  Member list (for groups)\n");
		printf("  ?  Print this list\n\n");
		goto again;
		/* NOTREACHED */
	case 't' :
		change_field("Title", TITLE, Person.title, dn);
		break;
	case 'e' :
		change_field("E-mail address", E_MAIL_ADDRESS, Person.email, dn);
		break;
	case 'p' :
		change_field("Home phone", HOME_PHONE, Person.hphon, dn);
		break;
	case 'P' :
		change_field("Business phone", PHONE_NUMBER, Person.bphon, dn);
		break;
	case 'a' :
		change_field("Home address", HOME_ADDRESS, Person.haddr, dn);
		break;
	case 'A' :
		change_field("Business address", POSTAL_ADDRESS, Person.baddr, dn);
		break;
	case 'f' :
		change_field("Fax number", FAX, Person.faxxx, dn);
		break;
	case 'm' :
		change_field("Member list", MEMBERS, Person.mmbrs, dn);
		break;
	case 'd' :
		change_field("Description", MLDESCRIPTION, Person.mldsc, dn);
		break;
	case 'x' :
		set_pw(dn);
		break;
#ifdef UOFM
	case 'u' :
	case 'U' :
		set_updates(dn);
		break;
#endif
	case '\0' :
		putchar('\n');
		break;
	default :
		printf("  Unrecognizable option\n\n");
		goto again;
		/* NOTREACHED */
	}
	return;
}

void set_pw(who)
char *who;
{
	char *cp;
	static char new[513];
	extern char * mygetpass();
	extern char password[];
	static char value[MAX_VAL_SIZE];
	static char *what[2];

	/* initialize 'what' buffer for use by dxi_modify() */
	what[0] = value;
	what[1] = NULL;

	if (verbose) {
		printf("  You indicated that you wanted to change your password.\n");
		printf("  You will now be prompted for your old, existing password.\n");
		printf("  Then you will be prompted twice for your new password.\n");
		printf("  If they match, and your old password is correct, your\n");
		printf("  password will be changed.\n\n");
	}
	cp = mygetpass("  Old password? ");
	if (strcmp(cp, password)) {
		printf("  Incorrect password -- not changed.\n");
		return;
	}
	cp = mygetpass("  New password? ");
	(void) strcpy(new, cp);
	cp = mygetpass("  Enter new password again to confirm: ");
	if (strcmp(cp, new)) {
		printf("  Password mismatch -- not changed.\n");
		return;
	}
	/* set the password here */
	sprintf(value, "%s=%s", PASSWORD, cp);
	if (dxi_modify(who, what) < 0) {
		dxi_perror("\tdxi_modify");
		return;
	}
	else if (verbose)
		printf("  Password changed.\n");
	(void) strcpy(password, cp);
	return;
}


/* generic routine for changing any field */
change_field(tag, id, val, who)
char *tag;			/* user-friendly tag identifying the field */
char *id;			/* X.500 internal name for the field */
char val[][MAX_VAL_SIZE];	/* current set of values for this field */
char *who;			/* name of user whose entry we are changing */
{
				
	char *cp;			/* miscellaneous pointer */
	char *get_value();		/* routine to extract values */
	static char buf[4 * RESP_SIZE];	/* for printing things */
	static char resp[RESP_SIZE];	/* for user input */
	static char value[MAX_VAL_SIZE];/* new value to be added */
	static char *what[2];		/* attribute to modify */

	/* initialize 'what' buffer for use by dxi_modify() */
	what[0] = value;
	what[1] = NULL;

	/*
	 *  If there is no current value associated with the attribute,
	 *  then this is the easy case.  Collect one (or more) attributes
	 *  from the user, and then call dxi_modify() to write the changes
	 *  to the DIXIE server.
	 */
	if (val[0][0] == '\0') {
		printf("\n  No current \"%s\"\n", tag);
		cp = get_value(id, "Enter a value");
		sprintf(value, "%s=%s", id, cp);
		for (;;) {
			printf("  Do you wish to add an additional value? ");
			fflush(stdout);
			fetch_buffer(resp, sizeof(resp), stdin);
			if ((resp[0] == 'y') || (resp[0] == 'Y')) {
				cp = get_value(id, "Enter an additional value");
				strcat(value, " & ");
				strcat(value, cp);
			}
			else
				break;
		}
		if (dxi_modify(who, what) < 0) {
			dxi_perror("\tdxi_modify");
			return;
		}
		else if (verbose)
			printf("  Modification of '%s' complete.\n", tag);
	}
	/*
	 *  There are values for this attribute already.  In this case,
	 *  we want to allow the user to delete all existing values,
	 *  add additional values to the ones there already, or just
	 *  delete some of the values already present.  DIXIE does not
	 *  handle modifications where the attribute occurs on the LHS
	 *  more than once.  So to delete entries and add entries, we
	 *  need to call dxi_modify() twice.
	 */
	else {
		sprintf(buf, "%s:  ", tag);
		print_entry(val, (char *) buf);
		printf("\n  Add a value, Delete a value, or Clear all values? ");
		fflush(stdout);
		fetch_buffer(resp, sizeof(resp), stdin);
		if ((resp[0] == 'a') || (resp[0] == 'A')) {
			cp = get_value(id, "Enter the value you wish to add");
			if (*cp == '\0') {
				if (verbose)
					printf("  No changes made.\n");
				return;
			}
			sprintf(value, "%s+=%s", id, cp);
			for (;;) {
				printf("  Add an additional value? ");
				fflush(stdout);
				fetch_buffer(resp, sizeof(resp), stdin);
				if ((resp[0] == 'y') || (resp[0] == 'Y')) {
					cp = get_value(id, "Enter another value you wish to add");
					strcat(value, " & ");
					strcat(value, cp);
				}
				else
					break;
			}
			if (dxi_debug)
				printf("Adding values {%s}\n", value);
			if (dxi_modify(who, what) < 0) {
				dxi_perror("  dxi_modify");
				return;
			}
			else if (verbose)
				printf("  Additions to '%s' complete.\n", tag);
		}
		else if ((resp[0] == 'd') || (resp[0] == 'D')) {
			if (verbose && (!strcmp(id, HOME_ADDRESS) || !strcmp(id, MLDESCRIPTION) || !strcmp(id, POSTAL_ADDRESS))) {
				printf("  WARNING!\n");
				printf("  You cannot delete multi-line values in this way.\n");
				printf("  You must clear the field, and then enter new values.\n\n");
			}
			cp = get_value(id, "Enter a value you wish to delete");
			if (*cp == '\0') {
				if (verbose)
					printf("  No changes made.\n");
				return;
			}
			sprintf(value, "%s-=%s", id, cp);
			if (dxi_debug)
				printf("Deleting value {%s}\n", cp);
			if (dxi_modify(who, what) < 0) {
				if (dxi_errno == DXI_NEX_ATTRVAL)
					printf("  Value incorrect -- could not find '%s'\n", cp);
				else
					dxi_perror("\tdxi_modify");
				return;
			}
			else if (verbose)
				printf("  Deletion of '%s' complete.\n", cp);
			for (;;) {
				printf("  Delete additional values? ");
				fflush(stdout);
				fetch_buffer(resp, sizeof(resp), stdin);
				if ((resp[0] == 'y') || (resp[0] == 'Y')) {
					cp = get_value(id, "Enter another value you wish to delete");
					sprintf(value, "%s-=%s", id, cp);
					if (dxi_debug)
						printf("Deleting value {%s}\n", cp);
					if (dxi_modify(who, what) < 0) {
						if (dxi_errno == DXI_NEX_ATTRVAL)
							printf("  Value incorrect -- could not find '%s'\n", cp);
						else
							dxi_perror("  dxi_modify");
						return;
					}
					else if (verbose)
						printf("  Deletion of '%s' complete.\n", cp);
				}
				else
					break;
			}
		}
		else if ((resp[0] == 'c') || (resp[0] == 'C')) {
			sprintf(value, "%s", id);
			if (dxi_debug)
				printf("Clearing attribute '%s'\n", id);
			if (dxi_modify(who, what) < 0) {
				dxi_perror("  dxi_modify");
				return;
			}
			else if (verbose)
				printf("  '%s' has been cleared.\n", tag);
		}
		else {
			if (verbose)
				printf("  No changes made.\n");
		}
	}
	return;
}

char *get_value(id, prompt)
char *id, *prompt;
{
	int count;		/* line # of new value -- if multiline */
	int needs_T61 = 0;	/* 1 if value can have T.61 characters */
	int multiline = 0;	/* 1 if this value is multiline */
	static char line[LINE_SIZE];	/* raw line from user */
	static char buffer[MAX_LINES * LINE_SIZE];	/* holds ALL of the 
							   lines we get */

	/*
	 *  To start with, we have one line of input from the user.
	 *  We should also zero out the buffer we'll use if we need to
	 *  perform any T.61 transformations on the input buffer stored
	 *  in 'line.'
	 *
	 *  Addresses and multiline description can span multiple lines.
	 *  Other attributes may not.
	 *
	 *  Addresses can contain T.61 characters.  We don't both with
	 *  other kinds of attributes.
	 */
	count = 1;
	(void) memset(buffer, 0, sizeof(buffer));
	if (!strcmp(id, POSTAL_ADDRESS) || !strcmp(id, HOME_ADDRESS)) {
		needs_T61 = 1;
		multiline = 1;
	}
	else if (!strcmp(id, MLDESCRIPTION))
		multiline = 1;
	printf("\n  %s:\n", prompt);

	/* fetch lines */
	for (;;) {
		if (multiline)
			printf(" %1d: ", count);
		else
			printf("  > ");
		fflush(stdout);
		fetch_buffer(line, sizeof(line), stdin);
		if (line[0] == '\0')
			break;

		/*
		 *  If this flag is set, this means that this is an
		 *  attribute which can contain T.61 characters.
		 *
		 *  Scan through the line of input supplied by the user
		 *  copying characters from 'line' to 'line2'.  If we
		 *  happen upon a character which requires a T.61
		 *  transformation, copy in a '{T.61}' into 'line2,'
		 *  and then copy in the two characters which represent
		 *  the ASCII hex value for that character.
		 */
		if (needs_T61) {
			int do_it = 0;
			static char line2[BUFSIZ];
			register char *p1, *p2;
	
			for (p1 = line ; *p1 != '\0'; p1++) {
				if (isT61(*p1)) {
					do_it = 1;
					break;
				}
			}
			if (do_it) {
				static char map[] = "0123456789abcdef";

				p1 = line;
				sprintf(line2, "{T.61}");
				p2 = line2 + strlen(line2);
				while (*p1 != '\0') {
					if (isT61(*p1)) {
						*p2++ = '\\';
						*p2++ = map[((*p1>>4) & 0x0f)];
						*p2++ = map[*p1 & 0x0f];
						p1++;
					}
					else
						*p2++ = *p1++;
				}
				(void) strcpy(line, line2);
			}
		}

		/*
		 *  Separate lines of multiline attribute values with
		 *  dollar signs.  Copy this line into the buffer we
		 *  use to collect up all of the user-supplied input
		 *  lines.  If this is not a multiline attribute, we
		 *  are done.
		 */
		if (count++ > 1)
			(void) strcat(buffer, "$");
		(void) strcat(buffer, line);
		if (!multiline)
			break;
	}
	if (dxi_debug)
		printf("  Value is [%s]\n", buffer);
	return(buffer);
}

#ifdef UOFM

void set_updates(who)
char *who;
{
	char *cp, *s;
	static char *what[2];
	extern char * fetch_update_setting();
	static char value[MAX_VAL_SIZE];
	static char response[16];

	/* initialize 'what' buffer for use by dxi_modify() */
	what[0] = value;
	what[1] = NULL;

	/* explain what the implications are */
	if (verbose) {
		printf("\n  By default, updates that are received from the Personnel\n");
		printf("  Office and the Office of the Registrar are applied to all\n");
		printf("  entries in the X.500 database each month.  Sometimes this\n");
		printf("  feature is undesirable.  For example, if you maintain your\n");
		printf("  entry in the X.500 database manually, you may not want to\n");
		printf("  have these updates applied to your entry, possibly overwriting\n");
		printf("  correct information with out-dated information.\n\n");
	}

	/* fetch the current setting */
	if ((cp = fetch_update_setting(who)) == NULL)
		return;
	if (!strcmp(cp, "TRUE"))
		printf("  Automatic updates are currently turned OFF\n");
	else if (!strcmp(cp, "FALSE"))
		printf("  Automatic updates are currently turned ON\n");
	else {
		fprintf(stderr, "  Unknown update flag -> [%s]\n", cp);
		return;
	}

	/* see if they want to change it */
	printf("\n  Change this setting [no]? ");
	fflush(stdout);
	(void) fgets(response, sizeof(response), stdin);
	for (s = response; isspace(*s); s++)
			;
	if ((*s == 'y') || (*s == 'Y')) {
		if (!strcmp(cp, "TRUE"))
			sprintf(value, "%s=FALSE", UPDATES);
		else
			sprintf(value, "%s=TRUE", UPDATES);
		if (dxi_modify(who, what) < 0) {
			dxi_perror("\tdxi_modify");
			return;
		}
		else if (verbose)
			printf("  Setting has been changed\n");
		return;
	}
	if (verbose)
			printf("  Setting has not been changed\n");
}

#endif
